// cl_zeta().

// General includes.
#include "cl_sysdep.h"

// Specification.
#include "cl_R_tran.h"


// Implementation.

#include "cl_lfloat.h"
#include "cl_LF_tran.h"
#include "cl_LF.h"
#include "cl_integer.h"
#include "cl_abort.h"
#include "cl_alloca.h"

cl_LF cl_zeta (int s, uintC len)
{
	if (!(s > 1))
		cl_abort();
	// Method:
	// zeta(s) = 1/(1-2^(1-s)) sum(n=0..infty, (-1)^n/(n+1)^s),
	// with convergence acceleration through exp(x), and evaluated
	// using the binary-splitting algorithm.
	var uintC actuallen = len+2; // 2 Schutz-Digits
	var uintL x = (uintL)(0.693148*intDsize*actuallen)+1;
	var uintL N = (uintL)(2.718281828*x);
	CL_ALLOCA_STACK;
	var pqd_series_term* args = (pqd_series_term*) cl_alloca(N*sizeof(pqd_series_term));
	var uintL n;
	for (n = 0; n < N; n++) {
		if (n==0) {
			new (&args[n].p) cl_I (1);
			new (&args[n].q) cl_I (1);
		} else {
			new (&args[n].p) cl_I (x);
			new (&args[n].q) cl_I (n);
		}
		new (&args[n].d) cl_I (evenp(n)
		                       ? expt_pos(n+1,s)
		                       : -expt_pos(n+1,s));
	}
	var cl_LF result = eval_pqd_series(N,args,actuallen);
	for (n = 0; n < N; n++) {
		args[n].p.~cl_I();
		args[n].q.~cl_I();
		args[n].d.~cl_I();
	}
	result = shorten(result,len); // verkrzen und fertig
	// Zum Schluss mit 2^(s-1)/(2^(s-1)-1) multiplizieren:
	return scale_float(result,s-1) / (ash(1,s-1)-1);
}
// Bit complexity (N = len): O(log(N)^2*M(N)).
